import logging
from spyql.parser import parse
from spyql.processor import Processor
import spyql.log


class Query:
    '''
    A SPyQL query than can be executed on top of a file or variables producing a file or a :class:`~spyql.query_result.QueryResult`.
    Example::

        query = Query("""
            SELECT row.name as first_name, row.age
            FROM data
            WHERE row.age > 30
        """)

        result = query(data=[
            {"name": "Alice", "age": 20, "salary": 30.0},
            {"name": "Bob", "age": 30, "salary": 12.0},
            {"name": "Charles", "age": 40, "salary": 6.0},
            {"name": "Daniel", "age": 43, "salary": 0.40},
        ])

        ## result:
        # (
        #    {"first_name": "Charles", "age": 40},
        #    {"first_name": "Daniel", "age": 43},
        # )
    '''

    def __init__(
        self,
        query: str,
        input_options: dict = {},
        output_options: dict = {},
        unbuffered=False,
        warning_flag="default",
        verbose=0,
        default_to_clause="MEMORY",
    ) -> None:
        """
        Creates a ``Query`` object (does not execute the query).

        :param query: a spyql query
        :type query: str
        :param input_options: options to be passed to the input processor.
            e.g. ``{"delimiter": ","}``
        :type input_options: dict, optional
        :param output_options: options to be passed to the output writer.
            e.g. ``{"delimiter": ";", "header": False}``
        :type output_options: dict, optional
        :param unbuffered: forces output to be unbuffered.
        :type unbuffered: bool, optional
        :param warning_flag: set to "error" to turn warnings into errors
            (halting execution)
        :type warning_flag: str, optional
        :param verbose: set the verbosity level:
            -2 to supress errors and warnings;
            -1 to supress warnings;
            0 to only show errors and warnings (default);
            1 to show additional info messages;
            2 to show additional debug messages.
        :type verbose: int, optional
        """

        logging.basicConfig(level=(3 - verbose) * 10, format="%(message)s")
        spyql.log.error_on_warning = warning_flag == "error"

        self.query = query
        self.parsed, self.strings = parse(query, default_to_clause)
        self.output_options = output_options
        self.input_options = input_options
        self.unbuffered = unbuffered
        self.__stats = None

        spyql.log.user_debug_dict("Parsed query", self.parsed)
        spyql.log.user_debug_dict("Strings", self.strings.strings)

    def __repr__(self) -> str:
        return f'Query("{self.query}")'

    def __call__(self, **kwargs):
        """
        Executes the query.
        ``kwargs`` can take in multiple variables that are included in the
        scope of the query.
        """
        processor = None
        result = None
        self.__stats = None
        try:
            # make the processor
            processor = Processor.make_processor(
                self.parsed,
                self.strings,
                self.input_options,
            )
            result, self.__stats = processor.go(
                output_options=self.output_options,
                user_query_vars=kwargs,
            )
        finally:
            # makes sure files are closed
            if processor:
                processor.close()
                if processor.writer:
                    processor.writer.close()
        return result

    def stats(self):
        """
        Returns a dictionary with statistics about the query execution,
        namely the number of rows in the input and output.

        :rtype: dict
        """
        return self.__stats



import re
import random
import string

STRING_PLACEHOLDER_LEN = 12


class QuotesHandler:
    def __init__(self):
        self.strings = {}

    # replaces quoted strings by placeholders to make parsing easier
    # populates dictionary of placeholders and the strings they hold
    def extract_strings(self, query):
        res = []
        quotes = [
            r"\'(\'\'|\\.|[^'])*\'",
            r'\"(\"\"|\\.|[^"])*\"'  # ,
            #    r'\`(\`\`|\\.|[^`])*\`'
        ]

        spans = [(0, 0)]
        for quote in quotes:
            spans.extend([m.span() for m in re.finditer(quote, query)])
        spans.append((len(query), 0))

        self.strings = {}
        for i in range(len(spans) - 1):
            if i > 0:
                sid = "".join(
                    random.choice(string.ascii_letters)
                    for _ in range(STRING_PLACEHOLDER_LEN)
                )
                sid = f"__{sid}__"
                res.append(sid)
                self.strings[sid] = query[spans[i][0] + 1 : spans[i][1] - 1]

            res.append(query[spans[i][1] : spans[i + 1][0]])

        return "".join(res)

    @staticmethod
    def string_placeholder_re():
        return r"\_\_[a-zA-Z]{%d}\_\_" % (STRING_PLACEHOLDER_LEN)

    # replace string placeholders by their actual strings
    def put_strings_back(self, text, quote=True):
        if not isinstance(text, str):
            return text

        quote_char = '"' if quote else ""
        sids = {m.group(0) for m in re.finditer(self.string_placeholder_re(), text)}
        sids = sids.intersection(self.strings)  # eliminate false positives
        for sid in sids:
            text = text.replace(sid, f"{quote_char}{self.strings[sid]}{quote_char}")
        return text



from spyql.nulltype import NULL
import spyql.log


class qdict(dict):
    """
    A dictionary that supports :data:`~spyql.nulltype.NULL` where items can
    be accessed like attributes::

        mydict = qdict({
            "a": 1,
            "b": {
                "c": 2
            },
            "d": None
        })
        mydict.a    # returns 1, same as mydict["a"]
        mydict.z    # returns NULL whenever a key is not found
        mydict.b.c  # returns 2, neested dicts also support attribute access
        mydict.b.x  # returns NULL, neested dicts are null-safe too
        mydict.d    # returns NULL, Nones are converted to NULLs)
    """

    @staticmethod
    def __none2null(value):
        if type(value) is list:
            # TODO consider conversion of tuples/sets/etc
            return [NULL if x is None else x for x in value]
        return NULL if value is None else value

    @staticmethod
    def __none2null_dict(adic):
        # TODO: this should work with a list of pairs and not only with a dict
        return {k: qdict.__none2null(v) for k, v in adic.items()}

    def __init__(self, adic, dirty=True, **kwargs):
        # dirty option keeps None values in dict instead of converting to NULL
        self.update(adic if dirty else qdict.__none2null_dict(adic), **kwargs)
        self.__dict__["_dirty"] = dirty

    def __getitem__(self, key):
        try:
            item = dict.__getitem__(self, key)
            if type(item) is dict:
                # lazy convertion of dicts
                # TODO consider convertion of lists of dicts
                return qdict(item, self._dirty)
            if self._dirty:
                # lazy convertion
                return qdict.__none2null(item)
            return item
        except KeyError:
            return self.__missing__(key)

    def __getattr__(self, key):
        if (
            not key.startswith("__") or key in self
        ):  # because of special methods like __getstate__
            return self[key]
        else:
            raise AttributeError

    def __setattr__(self, key, value):
        self[key] = value

    def values(self):
        """
        Returns a tuple of the dict values.
        Attention: does not return a view like dict!
        """
        return tuple([qdict.__none2null(x) for x in super().values()])

    def items(self):
        """
        Returns a zip of keys and values.
        Attention: does not return a view like dict!
        """
        return zip(self.keys(), self.values())

    # returns NULL when key is not found
    def __missing__(self, key):
        spyql.log.user_warning4func("key not found", KeyError(key), key)
        return NULL

    def __hash__(self):
        # TODO make dict immutable
        import json

        # TODO check if this is sufficienly efficient...
        # This only needs to guarantee that two equivalent dicts have the same hash
        return hash(json.dumps(self, default=str, sort_keys=True))



import csv
import pickle
from tabulate import tabulate  # https://pypi.org/project/tabulate/
import asciichartpy as chart
from math import nan
import sys
import io

from spyql.log import user_debug, user_error, user_error
from spyql.nulltype import NULL
from spyql.qdict import qdict
from spyql.query_result import QueryResult
from spyql.utils import is_row_collapsable


class Writer:
    @staticmethod
    def output_writers():
        return {
            "JSON": JSONWriter,
            "ORJSON": ORJSONWriter,
            "CSV": CSVWriter,
            "MEMORY": MemoryWriter,
            "SPY": SpyWriter,
            "PRETTY": PrettyWriter,
            "SQL": SQLWriter,
            "PLOT": PlotWriter,
        }

    @staticmethod
    def make_writer(to_clause, output_options={}):
        """
        Factory for making an output writer based on the parsed query
        """
        try:
            writer_name = to_clause
            if not to_clause:  # not TO clause, defaults to CSV
                writer_name = "CSV"
                return CSVWriter(**output_options)
            elif isinstance(to_clause, dict):  # there's an output data writer
                writer_name = to_clause["name"]
                writer = Writer.output_writers()[writer_name.upper()]
                output_options.update(to_clause["kwargs"])
                return writer(*to_clause["args"], **output_options)
            else:
                user_error(
                    f"Unknown writer '{writer_name}'",
                    SyntaxError("Error parsing TO statement"),
                    writer_name,
                )
        except TypeError as e:
            user_error(f"Could not create '{writer_name}' writer", e)

    def __init__(self, path=None, unbuffered=False):
        user_debug(f"Loading writer {self.__class__.__name__}")
        self.header = []
        self.path = path
        try:
            self.outputfile = open(path, "w") if path else sys.stdout
            if unbuffered:
                self.outputfile = io.TextIOWrapper(
                    open(self.outputfile.fileno(), "wb", 0), write_through=True
                )
        except Exception as e:
            user_error(f"Could not open output file {path}", e)

    def close(self):
        if self.path:
            self.outputfile.close()

    def writeheader(self, header):
        self.header = header

    def writerow(self, row):
        raise NotImplementedError

    def writerows(self, rows):
        for r in rows:
            self.writerow(r)

    def flush(self):
        pass

    def result(self) -> QueryResult:
        """Gets query result, in case of writing to memory"""
        return None


class CSVWriter(Writer):
    def __init__(self, path=None, unbuffered=False, header=True, **options):
        super().__init__(path, unbuffered)
        self.header_on = header
        self.csv = csv.writer(self.outputfile, **options)

    def writeheader(self, header):
        if self.header_on:
            self.csv.writerow(header)

    def writerow(self, row):
        self.csv.writerow(row)

    def writerows(self, rows):
        self.csv.writerows(rows)


def json_default(x):
    return None if x is NULL else str(x)


class JSONWriter(Writer):
    def __init__(self, path=None, unbuffered=False, default=json_default, **options):
        import json

        super().__init__(path, unbuffered)
        self.encoder = json.JSONEncoder(default=default, **options)
        self.encoder.encode({"a": 1})  # test options

    def writerow(self, row):
        obj = (
            row[0]
            if is_row_collapsable(row, self.header)
            else dict(zip(self.header, row))
        )
        self.outputfile.write(self.encoder.encode(obj) + "\n")


class ORJSONWriter(Writer):
    def __init__(self, path=None, unbuffered=False, default=json_default, option=0):
        super().__init__(path, unbuffered)
        try:
            # currently the only supported JSON external library is orjson,
            # which must be installed separately (it is not a dependency)
            import orjson

            self.orjson = orjson
        except ModuleNotFoundError as e:
            user_error(
                "`orjson` module not found. You might need to install it",
                e,
                "pip3 install orjson",
            )
        self.default = default
        self.option = option | orjson.OPT_APPEND_NEWLINE

    def writerow(self, row):
        # TODO optimization: only call `is_row_collapsable` in the 1st row?
        obj = (
            row[0]
            if is_row_collapsable(row, self.header)
            else dict(zip(self.header, row))
        )
        self.outputfile.buffer.write(
            self.orjson.dumps(obj, default=self.default, option=self.option)
        )


class CollectWriter(Writer):
    """
    Abstract writer that collects all records into a (in-memory) list and dumps all
    the output records at the end.
    Child classes must implement the `dumprows` method.
    """

    def __init__(self, path=None, unbuffered=False):
        super().__init__(path, unbuffered)
        self.all_rows = []  # needs to store output in memory

    def transformvalue(self, value):
        return None if value is NULL else value

    def transformrow(self, row):
        return [self.transformvalue(val) for val in row]

    def writerow(self, row):
        self.all_rows.append(self.transformrow(row))  # accumulates

    def dumprows(self, rows):
        raise NotImplementedError

    def flush(self):
        if self.all_rows:
            self.dumprows(self.all_rows)  # dumps


class MemoryWriter(CollectWriter):
    def transformrow(self, row):
        if not self.all_rows:
            # makes decision to collapse based on the first row
            self.__colapse = is_row_collapsable(row, self.header)
        return row[0] if self.__colapse else qdict(zip(self.header, row))

    def result(self):
        return QueryResult(self.all_rows, self.header)

    def dumprows(self, rows):
        pass


class PrettyWriter(CollectWriter):
    def __init__(self, path=None, unbuffered=False, header=True, **options):
        super().__init__(path, unbuffered)
        tabulate([[1, 2, 3]], **options)  # test options
        self.header_on = header
        self.options = options

    def dumprows(self, rows):
        # TODO handle default tablefmt
        self.outputfile.write(
            tabulate(
                rows,
                self.header if self.header_on else [],
                **self.options,
            )
        )
        self.outputfile.write("\n")


class PlotWriter(CollectWriter):
    def __init__(self, path=None, unbuffered=False, header=True, height=20):
        super().__init__(path, unbuffered)
        self.header_on = header
        self.height = height

    def transformvalue(self, value):
        return nan if value is NULL or value is None else value

    def dumprows(self, rows):
        colors = [
            chart.cyan,
            chart.red,
            chart.magenta,
            chart.lightgray,
            chart.green,
            chart.blue,
        ]
        config = {"height": self.height, "colors": colors}

        # first transpose rows into cols
        cols = list(map(list, zip(*rows)))

        self.outputfile.write(chart.plot(cols, config))
        if self.header and self.header_on:
            self.outputfile.write("\n\nLegend: ")
            for i in range(len(self.header)):
                self.outputfile.write(
                    "\t"
                    + colors[i % len(colors)]
                    + "─── "
                    + self.header[i]
                    + chart.reset
                    + " "
                )
            self.outputfile.write("\n")


class SpyWriter(Writer):
    def __init__(self, path=None, unbuffered=False):
        super().__init__(path, unbuffered)

    @staticmethod
    def pack(row):
        return pickle.dumps(row).hex() + "\n"

    def writeheader(self, header):
        # TODO first line is a dict with list of cols, version, etc
        self.outputfile.write(self.pack(header))

    def writerow(self, row):
        self.outputfile.write(SpyWriter.pack(row))


class SQLWriter(Writer):
    def __init__(
        self, path=None, unbuffered=False, chunk_size=1000, table="table_name"
    ):
        super().__init__(path, unbuffered)
        self.chunk_size = chunk_size
        self.table_name = table
        self.chunk = []

    def writeheader(self, header):
        self.statement = (
            f'INSERT INTO "{self.table_name}"('
            + ",".join(['"{}"'.format(h) for h in header])
            + ") VALUES {};\n"
        )

    def writerow(self, row):
        self.chunk.append(
            "({})".format(
                ",".join(
                    [
                        str(v)
                        if isinstance(v, int) or isinstance(v, float)
                        else (
                            "NULL"
                            if v is NULL or v is None
                            else "'{}'".format(str(v).replace("'", "''"))
                        )
                        for v in row
                    ]
                )
            )
        )

        if len(self.chunk) >= self.chunk_size:
            self.writestatement()

    def writestatement(self):
        self.outputfile.write(self.statement.format(",".join(self.chunk)))
        self.chunk = []

    def flush(self):
        if self.chunk:  # write leftovers...
            self.writestatement()



# Defines the NULL datatype that mimics SQL's NULL
# Unlike None, which throws exceptions, operations with NULLs return NULL
# Allows getting items so that `a.get('b', NULL).get('c', NULL)`` returns NULL
#  when `b` does not exist, and `NULL[x]` returns NULL


class NullType:
    def __new__(cls):
        return NULL

    def __reduce__(self):
        return (NullType, ())

    def __copy__(self):
        return NULL

    def __deepcopy__(self, a):
        return NULL

    def __call__(self, default):
        pass

    def __repr__(self):
        return "NULL"

    def __str__(self):
        return ""

    def __hash__(self):
        return hash("NULL")

    def __bool__(self):
        return False

    def __lt__(self, other):
        return self

    def __le__(self, other):
        return self

    def __eq__(self, other):
        return self

    def __ne__(self, other):
        return self

    def __ge__(self, other):
        return self

    def __gt__(self, other):
        return self

    def __abs__(self):
        return self

    def __add__(self, other):
        return self

    def __and__(self, other):
        return self

    def __floordiv__(self, other):
        return self

    def __invert__(self):
        return self

    def __lshift__(self, other):
        return self

    def __mod__(self, other):
        return self

    def __mul__(self, other):
        return self

    def __matmul__(self, other):
        return self

    def __neg__(self):
        return self

    def __or__(self, other):
        return self

    def __pos__(self):
        return self

    def __pow__(self, other):
        return self

    def __rshift__(self, other):
        return self

    def __sub__(self, other):
        return self

    def __truediv__(self, other):
        return self

    def __xor__(self, other):
        return self

    def __contains__(self, other):
        return False

    def __delitem__(self, other):
        pass

    def __getitem__(self, other):
        return self

    def __setitem__(self, other, val):
        pass

    def __radd__(self, other):
        return self

    def __rand__(self, other):
        return self

    def __rfloordiv__(self, other):
        return self

    def __rlshift__(self, other):
        return self

    def __rmod__(self, other):
        return self

    def __rmul__(self, other):
        return self

    def __rmatmul__(self, other):
        return self

    def __ror__(self, other):
        return self

    def __rpow__(self, other):
        return self

    def __rrshift__(self, other):
        return self

    def __rsub__(self, other):
        return self

    def __rtruediv__(self, other):
        return self

    def __rxor__(self, other):
        return self

    def __iadd__(self, other):
        return self

    def __iand__(self, other):
        return self

    def __ifloordiv__(self, other):
        return self

    def __ilshift__(self, other):
        return self

    def __imod__(self, other):
        return self

    def __imul__(self, other):
        return self

    __array_priority__ = 10000

    def __imatmul__(self, other):
        return self

    def __ior__(self, other):
        return self

    def __ipow__(self, other):
        return self

    def __irshift__(self, other):
        return self

    def __isub__(self, other):
        return self

    def __itruediv__(self, other):
        return self

    def __ixor__(self, other):
        return self

    def __len__(self):
        return 0

    def __iter__(self):
        return [].__iter__()

    def __round__(self, ndigits=0):
        return self

    def __trunc__(self):
        return self

    def __floor__(self):
        return self

    def __ceil__(self):
        return self

    def get(self, *args, **kwargs):
        return self


# singleton
try:
    NULL
    # explanation here:
    # https://stackoverflow.com/questions/41048643/how-to-create-a-second-none-in-python-making-a-singleton-object-where-the-id-is
except NameError:
    NULL = object.__new__(NullType)
    Null = NULL  # alias
    null = NULL  # alias



from typing import Any


class QueryResult(tuple):
    """
    Result of a query that writes outputs to memory.
    Tuple of dictionaries with easy access of columns by name as attributes.

    Accessing the value of the `age` column in the first row::

        result[0].age
        result[0]["age"]

    Collecting the age for all rows as a tuple::

        result.age
        result.col("age")

    Collecting the age for a subset of rows as a tuple::

        result[1:3].age
        result[1:3].col("age")

    Collecting the value of the first column for all rows as a tuple::

        result.col(0)

    Iterating over rows::

        for row in result:
            print(row.age, row.another_column)

    """

    def __new__(cls, __values, __colnames):
        return super(QueryResult, cls).__new__(cls, __values)

    def __init__(self, __values, __colnames):
        self.__colnames = tuple(__colnames)

    def col(self, idx):
        """
        Collects and returns a tuple with all values of the col defined by
        `idx`.

        :param idx: if `idx` is an integer it refers to the nth column
            (0-based indexing). If `idx` is a string it refers to the name of
            the column.
        """
        if isinstance(idx, int):
            idx = self.__colnames[idx]
        if isinstance(idx, str):
            return tuple([r[idx] for r in self])
        return TypeError(f"Column index of type {type(idx)} not supported")

    def __getattr__(self, name: str) -> Any:
        return self.col(name)

    def __getitem__(self, idx):
        res = super().__getitem__(idx)
        return QueryResult(res, self.__colnames) if isinstance(res, tuple) else res

    def colnames(self):
        """Returns a tuple with the name of each column.

        :rtype: tuple[str]"""
        return self.__colnames



import json
from tempfile import gettempdir

from spyql.query import Query
from spyql.utils import join_paths
from spyql.nulltype import NULL
from spyql.qdict import qdict
import os
import math

raw_data = [
    {"name": "A", "age": 20, "salary": 30.0},
    {"name": "B", "age": 30, "salary": 12.0},
    {"name": "C", "age": 40, "salary": 6.0},
    {"name": "D", "age": 50, "salary": 0.40},
]


def make_json():
    json_fpath = join_paths(gettempdir(), "spyql_test.jsonl")
    with open(json_fpath, "w") as f:
        for d in raw_data:
            f.write(json.dumps(d) + "\n")
    return json_fpath


def make_csv():
    csv_fpath = join_paths(gettempdir(), "spyql_test.csv")
    with open(csv_fpath, "w") as f:
        f.write(
            """name, age, salary
    A, 20, 30.
    B, 30, 12.
    C, 40, 6.
    D, 50, 0.40"""
        )
    return csv_fpath


def test_query_result():
    query = Query(
        "SELECT row.name as first_name, row.age as user_age FROM data WHERE"
        " row.age > 30"
    )
    out = query(data=raw_data)
    assert out == (
        {"first_name": "C", "user_age": 40},
        {"first_name": "D", "user_age": 50},
    )
    assert out.first_name == ("C", "D")
    assert out.col("first_name") == ("C", "D")
    assert out.col(0) == ("C", "D")
    assert out[0:2].first_name == ("C", "D")
    assert out.first_name[0] == "C"
    assert out[0].first_name == "C"
    assert out.user_age == (40, 50)
    assert out.col("user_age") == (40, 50)
    assert out.col(1) == (40, 50)
    assert out[0:2].user_age == (40, 50)
    assert out.user_age[-1] == 50
    assert out[-1].user_age == 50
    assert out.colnames() == ("first_name", "user_age")
    assert out.I_do_not_exist == (NULL, NULL)
    assert out[0].I_do_not_exist is NULL
    try:
        out.col(1000)
        assert False
    except IndexError:
        assert True
    assert query.stats() == {"rows_in": 4, "rows_out": 2}


def test_json_read():
    json_fpath = make_json()
    query = Query(
        "SELECT json.name as first_name, json.age as user_age FROM"
        f" json('{json_fpath}') WHERE json.age > 30"
    )
    out = query()
    assert out == (
        {"first_name": "C", "user_age": 40},
        {"first_name": "D", "user_age": 50},
    )
    os.remove(json_fpath)


def test_csv_read():
    csv_fpath = make_csv()
    query = Query(
        f"SELECT name as first_name, age as user_age FROM csv('{csv_fpath}') WHERE age"
        " > 30"
    )
    out = query()
    assert out == (
        {"first_name": "C", "user_age": 40},
        {"first_name": "D", "user_age": 50},
    )
    os.remove(csv_fpath)


def test_csv_write():
    csv_fpath = make_csv()
    target_csv = join_paths(gettempdir(), "spyql_test_write.csv")
    query = Query(
        f"SELECT name, age FROM csv('{csv_fpath}') WHERE age > 30 TO"
        f" csv('{target_csv}')"
    )
    query()

    with open(target_csv, "r") as f:
        out = f.read()

    assert out.strip().replace("\n", " ") == "name,age C,40 D,50"

    os.remove(csv_fpath)
    os.remove(target_csv)


def read_json(filename):
    with open(filename, "r") as f:
        return [json.loads(l) for l in f.read().strip().splitlines()]


def test_json_write():
    csv_fpath = make_csv()
    target_json = join_paths(gettempdir(), "spyql_test_write1.jsonl")
    Query(
        f"SELECT name, age FROM csv('{csv_fpath}') WHERE age > 30 TO"
        f" json('{target_json}')"
    )()
    assert read_json(target_json) == [
        {"name": "C", "age": 40},
        {"name": "D", "age": 50},
    ]
    os.remove(csv_fpath)
    os.remove(target_json)


def test_csv_read_json_write():
    csv_fpath = make_csv()
    target_json = join_paths(gettempdir(), "spyql_test_write2.jsonl")
    Query(
        f"SELECT name, age FROM csv('{csv_fpath}') WHERE age > 30 TO"
        f" json('{target_json}')"
    )()
    assert read_json(target_json) == [
        {"name": "C", "age": 40},
        {"name": "D", "age": 50},
    ]
    os.remove(csv_fpath)
    os.remove(target_json)


def test_read_invalid_file():
    try:
        Query("SELECT * FROM csv('where is this file???')")()
        assert False
    except FileNotFoundError:
        assert True


def test_write_invalid_file():
    try:
        Query("SELECT * FROM range(10) TO csv('this * is a bad / file name')")()
        assert False
    except FileNotFoundError:
        assert True


def test_equi_join():
    query = Query("SELECT row.name, names[row.name] AS ext_name FROM data")
    out = query(data=raw_data, names=qdict({"A": "Alice", "C": "Chris", "D": "Daniel"}))
    assert out == (
        {"name": "A", "ext_name": "Alice"},
        {"name": "B", "ext_name": NULL},
        {"name": "C", "ext_name": "Chris"},
        {"name": "D", "ext_name": "Daniel"},
    )


def test_globals():
    # accessing `os`` module and `raw_data` via `globals()`
    query = Query(
        "SELECT row.name as first_name, row.age as user_age, os.getpid() > 0 as test_os"
        " FROM raw_data WHERE row.age > 30"
    )
    out = query(**globals())
    assert out == (
        {"first_name": "C", "user_age": 40, "test_os": True},
        {"first_name": "D", "user_age": 50, "test_os": True},
    )


def test_readme():
    # TODO test all recipes in the README

    query = Query(
        'IMPORT hashlib as hl SELECT hl.md5(col1.encode("utf-8")).hexdigest() as h FROM'
        " data"
    )
    out = query(data=["a", "b", "c"])
    assert out == (
        {"h": "0cc175b9c0f1b6a831c399e269772661"},
        {"h": "92eb5ffee6ae2fec3ad71c777531578f"},
        {"h": "4a8a08f09d37b73795649038408b5f33"},
    )
    assert query.stats() == {"rows_in": 3, "rows_out": 3}

    query = Query(
        "SELECT row.invoice_num AS id, row.items_sold.name AS name,"
        " row.items_sold.price AS price FROM data EXPLODE row.items_sold"
    )
    out = query(
        data=[
            {
                "invoice_num": 1028,
                "items_sold": [
                    {"name": "tomatoes", "price": 1.5},
                    {"name": "bananas", "price": 2.0},
                ],
            },
            {"invoice_num": 1029, "items_sold": [{"name": "peaches", "price": 3.12}]},
        ]
    )

    assert out == (
        {"id": 1028, "name": "tomatoes", "price": 1.5},
        {"id": 1028, "name": "bananas", "price": 2.0},
        {"id": 1029, "name": "peaches", "price": 3.12},
    )

    expectation = tuple(
        [10 * math.cos(col1 * ((math.pi * 4) / 90)) for col1 in range(80)]
    )
    out = Query("SELECT 10 * cos(col1 * ((pi * 4) / 90)) FROM range(80)")()
    assert out.col(0) == expectation
    assert len(out.colnames()) == 1

    out = Query("SELECT * FROM [10 * cos(i * ((pi * 4) / 90)) for i in range(80)]")()
    assert out.col(0) == expectation
    assert len(out.colnames()) == 1



from spyql.quotes_handler import QuotesHandler
import spyql.utils
import spyql.log
from spyql.processor import Processor
from spyql.writer import Writer
import re

query_struct_keywords = [
    "import",
    "select",
    "from",
    "explode",
    "where",
    "group by",
    "order by",
    "limit",
    "offset",
    "to",
]


def get_agg_funcs():
    import spyql.agg
    import inspect

    # TODO replace this by register mechanism to allow for user-defined aggregation
    funcs = inspect.getmembers(spyql.agg, inspect.isfunction)
    return {f[0] for f in funcs}


agg_funcs = get_agg_funcs()


def extract_funcs(expr):
    return re.findall(r"(\w+)\s*\(", expr)


def has_agg_func(expr):
    return agg_funcs.intersection(extract_funcs(expr))


def throw_error_if_has_agg_func(expr, clause_name):
    a = has_agg_func(expr)
    if has_agg_func(expr):
        spyql.log.user_error(
            f"aggregate functions are not allowed in {clause_name} clause",
            SyntaxError("bad query"),
            ",".join(a),
        )


def has_reference2row(expr):
    return re.search(r"\brow\b", expr) is not None


def clean_query(q):
    """makes sure that queries start with a space (required for parse_structure)"""
    q = " " + q.strip()
    return q


def parse_structure(q):
    """parse the supported keywords, which must follow a given order"""
    keys = query_struct_keywords
    last_pos = 0
    key_matches = []
    for key in keys:
        entry = re.compile(fr"\s+{key}\s+".replace(" ", r"\s+"), re.IGNORECASE).search(
            q, last_pos
        )
        if entry:
            entry = entry.span()
            last_pos = entry[1]
        key_matches.append(entry)

    # # Alternative code where order is not enforced:
    # key_matches = [re.search(fr"\s+{key}\s+", q, re.IGNORECASE) for key in keys]
    # key_matches = [(m.span() if m else None)  for m in key_matches]

    d = {}
    for i in range(len(query_struct_keywords)):
        if not key_matches[i]:
            d[keys[i]] = None
            continue
        st = key_matches[i][1]
        nd = len(q)
        for j in range(i + 1, len(keys)):
            if key_matches[j]:
                nd = key_matches[j][0]
                break

        # this list should be empty, otherwise order of clauses was not respected
        misplaced_keys = list(
            filter(
                None,
                [
                    re.compile(
                        fr"\s+{k}\s+".replace(" ", r"\s+"), re.IGNORECASE
                    ).search(q[st:nd])
                    for k in keys
                ],
            )
        )
        if misplaced_keys:
            spyql.log.user_error(
                "could not parse query",
                SyntaxError(f"misplaced '{misplaced_keys[0][0].strip()}' clause"),
            )
        d[keys[i]] = q[st:nd]

    return d


def pythonize(s):
    """replaces sql/custom syntax by python syntax"""
    # TODO check for special SQL stuff such as in, is, like
    # s = re.compile(r"([^=<>])={1}([^=])").sub(r"\1==\2", s)
    # DECISION: expressions are PURE python code :-)

    # make sure the `as` keyword is always lowcase
    # (currently only needed for imports)
    s = re.compile(r"\s+AS\s+", re.IGNORECASE).sub(" as ", s)

    # replace count_agg(*) and count_distinct_agg(*) by appropriate calls
    s = re.compile(r"\bcount\_agg\s*\(\s*\*\s*\)").sub("count_agg(1)", s)
    s = re.compile(r"\bcount\_distinct\_agg\s*\(\s*\*\s*\)").sub(
        "count_distinct_agg(tuple(_values))", s
    )

    # easy shortcut for navigating through dics (of dics)
    # e.g.   `json->hello->'planet hearth'` converts into
    #       `json['hello']['planet hearth']`

    # first replace quoted keys (they do not need quotes)
    s = re.compile(r"->(%s)" % (QuotesHandler.string_placeholder_re())).sub(r"[\1]", s)
    # then replace unquoted keys (they need quotes)
    s = re.compile(r"->([^\d\W]\w*)").sub(r"['\1']", s)

    return s


def split_multi_expr_clause(s):
    """
    Transforms "abc, (123 + 1) * 2, f(a,b)"
    into ["abc", "(123 + 1) * 2", "f(a,b)"]
    """
    sin = list(s)
    sep = [-1]
    rb = 0  # ()
    cb = 0  # {}
    sb = 0  # []
    for i in range(len(sin)):
        c = sin[i]
        if c == "(":
            rb = rb + 1
        elif c == ")":
            rb = rb - 1
        elif c == "{":
            cb = cb + 1
        elif c == "}":
            cb = cb - 1
        elif c == "[":
            sb = sb + 1
        elif c == "]":
            sb = sb - 1
        elif c == "," and rb == 0 and cb == 0 and sb == 0:
            sep.append(i)
    sep.append(None)
    parts = [s[sep[i] + 1 : sep[i + 1]].strip() for i in range(len(sep) - 1)]

    return parts


def parse_select(sel, strings):
    """splits the SELECT clause into columns and find their names"""
    # TODO support column alias without AS
    modif_pattern = r"^\s*(?:(DISTINCT)\s+)?(?:(PARTIALS)\s+)?"
    modifs = re.search(modif_pattern, sel.upper())
    has_distinct = "DISTINCT" in modifs.groups()
    has_partials = "PARTIALS" in modifs.groups()
    sel = sel[modifs.span()[1] :]  # remove modifiers from expression

    res = []
    as_pattern = re.compile(r"\s+AS\s+", re.IGNORECASE)
    for expr in split_multi_expr_clause(sel):
        sas = re.search(as_pattern, expr)
        name = ""
        if sas:
            name = expr[(sas.span()[1]) :].strip()
            expr = expr[: (sas.span()[0])]
        else:
            # automatic output column name from expression
            # removes json/row 'variables' reference (visual garbage)
            name = re.compile(r"(\b)(?:json|row)(->|\[|\.)").sub(r"\1", expr)
            # makes the string a valid python variable name
            name = spyql.utils.make_str_valid_varname(strings.put_strings_back(name))

        if expr.strip() == "*":
            expr = "*"
            name = "*"
        else:
            name = strings.put_strings_back(name, quote=False)
            expr = f"{make_expr_ready(expr, strings)}"

        res.append({"name": name, "expr": expr})

    return res, has_distinct, has_partials


def parse_orderby(clause, strings):
    """splits the ORDER BY clause and handles modifiers"""

    res = []
    mod_pattern = re.compile(r"(?:\s+(DESC|ASC))?(?:\s+NULLS\s+(FIRST|LAST)\s*)?$")
    for expr in split_multi_expr_clause(clause):
        modifs = re.search(mod_pattern, expr.upper())
        rev = "DESC" in modifs.groups()
        rev_nulls = ((not rev) and "FIRST" in modifs.groups()) or (
            rev and "LAST" in modifs.groups()
        )
        expr = expr[: (modifs.span()[0])]  # remove modifiers
        try:
            expr = int(expr)  # special case: expression is output column number
        except ValueError:
            expr = make_expr_ready(expr, strings)

        res.append({"expr": expr, "rev": rev, "rev_nulls": rev_nulls})

    return res


def parse_groupby(clause, select, strings):
    """splits the GROUP BY clause"""

    res = []
    for expr in split_multi_expr_clause(clause):
        try:
            expr = int(expr)  # special case: expression is output column number
            # in the case of group by, the expression is copied from select to avoid
            # group by depending on select (see spyql.processor._go)
            expr = select[expr - 1]["expr"]
        except ValueError:
            expr = make_expr_ready(expr, strings)
        throw_error_if_has_agg_func(expr, "GROUP BY")
        res.append({"expr": expr})

    return res


def extract_args(*args, **kwargs):
    return {"args": args, "kwargs": kwargs}


def parse_fromto(clause, strings, formats_list):
    formats_pattern = re.compile(rf"^(\w+)\s*(\(?.*$)")
    sformats = re.search(formats_pattern, clause)
    if not sformats or sformats.groups()[0].upper() not in formats_list:
        return make_expr_ready(clause, strings)  # python expression
    sformats = sformats.groups()
    res = eval(
        "extract_args"
        + (
            strings.put_strings_back(make_expr_ready(sformats[1], strings), strings)
            if sformats[1]
            else "()"
        )
    )
    res["name"] = sformats[0]
    return res


def make_expr_ready(expr, strings):
    return pythonize(expr).strip()


def parse(query, default_to_clause="MEMORY"):
    """parses the spyql query"""
    query = clean_query(query)
    strings = QuotesHandler()
    query = strings.extract_strings(query)
    query_has_agg_funcs = has_agg_func(query)
    prs = parse_structure(query)
    prs["hints"] = {"has_reference2row": has_reference2row(query)}
    if not prs["to"]:
        prs["to"] = default_to_clause

    if not prs["select"]:
        spyql.log.user_error(
            "could not parse query", SyntaxError("SELECT keyword is missing")
        )

    prs["select"], prs["distinct"], prs["partials"] = parse_select(
        prs["select"], strings
    )

    for clause in set(query_struct_keywords) - {
        "select",
        "limit",
        "offset",
        "group by",
        "order by",
    }:
        if prs[clause]:
            if clause in {"where", "from"}:
                throw_error_if_has_agg_func(prs[clause], clause.upper())
            if clause == "from":
                prs[clause] = parse_fromto(
                    prs[clause], strings, Processor.input_processors().keys()
                )
            elif clause == "to":
                prs[clause] = parse_fromto(
                    prs[clause], strings, Writer.output_writers().keys()
                )
            else:
                prs[clause] = make_expr_ready(prs[clause], strings)

    for clause in {"group by"}:
        if prs[clause]:
            prs[clause] = parse_groupby(prs[clause], prs["select"], strings)
        elif query_has_agg_funcs:
            # creates a dummy group by with a constant if there are agg functions
            # e.g. `select count_agg(*) from csv`
            prs[clause] = [{"expr": "'_OVERALL_'"}]
            if prs["order by"]:
                spyql.log.user_warning(
                    "ORDER BY is useless since output will have a single result"
                )
        if prs[clause] and prs["distinct"]:
            # This is feasible to implement but currently not supported
            spyql.log.user_error(
                "DISTINCT cannot be used in aggregation queries",
                SyntaxError("bad query"),
            )

    for clause in {"order by"}:
        if prs[clause]:
            prs[clause] = parse_orderby(prs[clause], strings)

    for clause in {"limit", "offset"}:
        if prs[clause]:
            try:
                val = int(prs[clause])
                prs[clause] = val if val > 0 else 0
            except ValueError:
                prs[clause] = None

    return (prs, strings)



from spyql.nulltype import NULL
import spyql.log

# functions that support NULLs (and that need to be replaced in the query)
NULL_SAFE_FUNCS = {
    "int": "int_",
    "float": "float_",
    "str": "str_",
    "complex": "complex_",
}


# returns default if val is NULL otherwise returns val
def coalesce(val, default):
    if val is NULL:
        return default
    return val


ifnull = coalesce  # alias


# returns NULL if a equals b otherwise returns a
def nullif(a, b):
    if a == b:
        return NULL
    return a


# returns NULL if any argument equals NULL
def null_safe_call(fun, *args, **kwargs):
    if NULL in args or NULL in kwargs.values():
        return NULL
    return fun(*args, **kwargs)


# NULL-safe functions
def float_(a):
    if a is NULL:
        return NULL
    try:
        return float(a)
    except ValueError as e:
        spyql.log.conversion_warning("float", e, a)
        return NULL


def int_(a, *args, **kwargs):
    if a is NULL or NULL in args or NULL in kwargs.values():
        return NULL
    try:
        return int(a, *args, **kwargs)
    except ValueError as e:
        spyql.log.conversion_warning("int", e, a, **kwargs)
        return NULL


def complex_(*args):
    if NULL in args:
        return NULL
    try:
        return complex(*args)
    except ValueError as e:
        spyql.log.conversion_warning("complex", e, *args)
        return NULL


def str_(*args, **kwargs):
    if NULL in args or NULL in kwargs.values():
        return NULL
    return str(*args, **kwargs)



from spyql.query import Query
import spyql.utils
import spyql.log
import click


def parse_options(ctx, param, options):
    options = [opt.split("=", 1) for opt in options]
    for opt in options:
        if len(opt) < 2:
            raise click.BadParameter(
                f"bad format for option '{opt[0]}', format must be 'option=value'"
            )
    return {kv[0]: spyql.utils.try2eval(kv[1], globals()) for kv in options}


@click.command()
@click.argument("query")
@click.option(
    "-I",
    "input_opt",
    type=click.UNPROCESSED,
    callback=parse_options,
    multiple=True,
    help=(
        "Set input options in the format 'option=value'. Example: -Idelimiter=,"
        " -Iheader=False"
    ),
)
@click.option(
    "-O",
    "output_opt",
    type=click.UNPROCESSED,
    callback=parse_options,
    multiple=True,
    help=(
        "Set output options in the format 'option=value'. Example: -Odelimiter=,"
        " -Oheader=False"
    ),
)
@click.option(
    "--unbuffered",
    "-u",
    is_flag=True,
    help="Force output to be unbuffered.",
)
@click.option(
    "--verbose",
    "-v",
    "verbose",
    default=0,
    help=(
        "Set verbose level: -2 to supress errors and warnings; -1 to supress warnings;"
        " 0 to only show errors and warnings (default); 1 to show additional info"
        " messages; 2 to show additional debug messages."
    ),
)
@click.option(
    "-W",
    "warning_flag",
    type=click.Choice(["default", "error"]),
    default="default",
    help=(
        "Set if warnings are turned into errors or if warnings do not halt execution"
        " (default)."
    ),
)
@click.version_option(version=spyql.__version__)
def main(query, warning_flag, verbose, unbuffered, input_opt, output_opt):
    """
    Tool to run a SpyQL QUERY over text data.
    For more info visit: https://github.com/dcmoura/spyql

    \b
    [ IMPORT python_module [ AS identifier ] [, ...] ]
    SELECT [ DISTINCT | PARTIALS ]
        [ * | python_expression [ AS output_column_name ] [, ...] ]
        [ FROM csv | spy | text | python_expression | orjson | json [ EXPLODE path ] ]
        [ WHERE python_expression ]
        [ GROUP BY output_column_number | python_expression  [, ...] ]
        [ ORDER BY output_column_number | python_expression
            [ ASC | DESC ] [ NULLS { FIRST | LAST } ] [, ...] ]
        [ LIMIT row_count ]
        [ OFFSET num_rows_to_skip ]
        [ TO csv | orjson | json | spy | sql | pretty | plot ]
    """

    out = Query(
        query,
        input_opt,
        output_opt,
        unbuffered,
        warning_flag,
        verbose,
        default_to_clause="CSV",
    )()

    spyql.log.user_info(f"Output Meta: {out}")


if __name__ == "__main__":
    main()
